/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: ChooseVar.c,v 0.9.1.2 1991/12/16 20:49:58 leei Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/copyright.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>

#include <vim/Popups.h>
#include <vim/ChooseVarP.h>
#include <vim/strstr.h>

static XrmQuark charQ;
static XrmQuark ucharQ;
static XrmQuark shortQ;
static XrmQuark ushortQ;
static XrmQuark intQ;
static XrmQuark uintQ;
static XrmQuark floatQ;
static XrmQuark doubleQ;

static XtResource resources[] = {
#define offset(field) XtOffset(ChooseVarWidget, chooseVar.field)
    /* {name, class, type, size, offset, default_type, default_addr}, */
    { XtNselectFont, XtCBoldFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(sel_font), XtRString,
	"-adobe-helvetica-bold-r-*-*-*-120-75-75-*-*-*-*" },
    { XtNunselectFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(unsel_font), XtRString,
	"-adobe-helvetica-medium-r-*-*-*-120-75-75-*-*-*-*" },
    { XtNchoices, XtCChoices, XtRChoices, sizeof(cvChoices),
	offset(choices), XtRImmediate, NULL },
    { XtNcvTitle, XtCTitle, XtRString, sizeof(String),
	offset(title), XtRString, "Choose Values" }
#undef offset
};

static void 
  ClassInitialize _DCL_VOID;
static void 
  Initialize _DCL((request, new),
		  ChooseVarWidget request _AND ChooseVarWidget new );
static Boolean
  SetValues _DCL((w, request, new),
		 ChooseVarWidget w _AND 
		 ChooseVarWidget request _AND
		 ChooseVarWidget new );

ChooseVarWidgetClassRec chooseVarWidgetClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass)(&transientShellClassRec),
    /* class_name		*/	"ChooseVar",
    /* widget_size		*/	sizeof(ChooseVarWidgetRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	XtInheritRealize,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	NULL,
    /* expose			*/	NULL,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* composite_class fields */
    /* geometry_manager   */    XtInheritGeometryManager,
    /* change_managed     */    XtInheritChangeManaged,
    /* insert_child       */    XtInheritInsertChild,
    /* delete_child       */    XtInheritDeleteChild,
    /* extension          */    NULL
  },{
    /* extension          */    NULL
  },{
    /* extension          */    NULL
  },{
    /* extension          */    NULL
  },{
    /* extension          */    NULL
  },{ /* chooseVar fields */
    /* empty			*/	0
  }
};

WidgetClass chooseVarWidgetClass = (WidgetClass)&chooseVarWidgetClassRec;

static void Setup _DCL((w), Widget w );

ChooseVarWidget
findCVWidget _IMPL((w), Widget w )
{
    while ( w && !XtIsSubclass(w, chooseVarWidgetClass) )
      w = XtParent(w);
    return (ChooseVarWidget) w;
}

static Widget
findShell _IMPL((w), Widget w )
{
    for ( ; w; w = w->core.parent )
      if ( XtIsShell(w) ) return w;

    return NULL;
}

static void ActionsInitialize _DCL((w), Widget w);

static void
Initialize _IMPL((request, new),
		 ChooseVarWidget request _AND ChooseVarWidget new )
{
    Widget shell = findShell(new->core.parent);
    Visual *vis;
    Colormap cmap;

    XtVaGetValues( shell, XtNvisual, &vis, XtNcolormap, &cmap, NULL );
    if ( vis != new->shell.visual ) new->shell.visual = vis;
    if ( cmap != new->core.colormap ) new->core.colormap = cmap;
    Setup((Widget)new);

    ActionsInitialize((Widget) new);
    XtInstallAcceleratorsBelow(new, new);
}

static Boolean
SetValues _IMPL((w, request, new),
		ChooseVarWidget w _AND 
		ChooseVarWidget request _AND 
		ChooseVarWidget new )
{
    ChooseVarWidgetPart 
      *oldch = &w->chooseVar,
      *reqch = &request->chooseVar,
      *newch = &new->chooseVar;
    Boolean resize = FALSE;

    return resize;
}


static int num_cvTypes = 0;
static struct cvTypes {
    XrmQuark type;
    unsigned char idx;
    Widget (*createf) __((Widget, cvChoice *));
    void (*initf) __((cvChoice *, Boolean));
    void (*resetf) __((cvChoice *));
    void (*changef) __((Widget));
    Boolean (*changedf) __((cvChoice *));
} *cvTypeList = NULL;
static int len_cvTypeList = 0;

static int
findType _IMPL((typeq, idx), XrmQuark typeq _AND int idx )
{
    int i;

    for ( i = 0; i < num_cvTypes; ++i )
      if ( cvTypeList[i].type == typeq && cvTypeList[i].idx == idx )
	return i;
    return -1;
}

Boolean
cvRegisterType _IMPL((type, idx, cf, initf, resetf, changef, changedf),
		     String type _AND int idx _AND 
		     Widget (*cf) __((Widget, cvChoice*)) _AND
		     void (*initf) __((cvChoice*, Boolean)) _AND
		     void (*resetf) __((cvChoice*)) _AND
		     void (*changef) __((Widget)) _AND
		     Boolean (*changedf) __((cvChoice*)) )
{
    XrmQuark typeq = XrmStringToQuark(type);
    int n = findType(typeq, idx);

    if ( n >= 0 ) {
	cvTypeList[num_cvTypes].createf = cf;
	cvTypeList[num_cvTypes].initf = initf;
	cvTypeList[num_cvTypes].resetf = resetf;
	cvTypeList[num_cvTypes].changef = changef;
	cvTypeList[num_cvTypes].changedf = changedf;
	return False;
    } else {
	if ( num_cvTypes >= len_cvTypeList ) {
	    len_cvTypeList += 10;
	    cvTypeList = (struct cvTypes *)
	      XtRealloc(cvTypeList, len_cvTypeList*sizeof(*cvTypeList));
	}

	cvTypeList[num_cvTypes].type = typeq;
	cvTypeList[num_cvTypes].idx = idx;
	cvTypeList[num_cvTypes].createf = cf;
	cvTypeList[num_cvTypes].initf = initf;
	cvTypeList[num_cvTypes].resetf = resetf;
	cvTypeList[num_cvTypes].changef = changef;
	cvTypeList[num_cvTypes].changedf = changedf;
	++num_cvTypes;
	return True;
    }
}

void
  cvSetChoices _IMPL((w, descr, num),
		     Widget w _AND cvChoiceDescr *descr _AND Cardinal num)
{
    ChooseVarWidgetPart *cvp = &((ChooseVarWidget)w)->chooseVar;
    int i, j;
    int prev_num = cvp->choices.num;
    cvChoice *prev = cvp->choices.ch;
    
    cvp->choices.num = num;
    cvp->choices.ch = (cvChoice *) XtCalloc(num, sizeof(cvChoice));

    for ( i = 0; i < num; ++i, ++descr ) {
	/* See if the associated entry already exists. */
	for ( j = 0; j < prev_num; ++j ) {
	    if ( prev[j].name && !strcmp(prev[j].name, descr->name) )
	      break;
	}

	/* If it does, use it; otherwise, initialize a new one. */
	if ( j < prev_num ) {
	    cvp->choices.ch[i] = prev[j];	/* Copy it. */
	    prev[j].name = 0;			/* Mark against cleanup. */
	} else {
	    bzero(&cvp->choices.ch[i], sizeof(cvChoice));
	    cvp->choices.ch[i].name = descr->name;
	    cvp->choices.ch[i].type = XrmStringToQuark(descr->type);
	    cvp->choices.ch[i].idx = descr->idx;
	    cvp->choices.ch[i].prompt = descr->prompt;
	    cvp->choices.ch[i].var = descr->var;
	    cvp->choices.ch[i].changep = descr->changep;
	    cvp->choices.ch[i].args = descr->args;
	    cvp->choices.ch[i].cv_widget = w;
	}
    }

    /* Now cleanup the previous entries. */
    for ( j = 0; j < prev_num; ++j ) {
	if ( prev[j].name ) {
	    if ( prev[j].widget ) XtDestroyWidget(prev[j].widget);
	    if ( prev[j].cv_data ) XtFree(prev[j].cv_data);
	}
    }
    if ( prev_num > 0 ) XtFree(prev);

    Setup(w);
}

static void
  Setup _IMPL((w), Widget w )
{
    ChooseVarWidgetPart *cvp = &((ChooseVarWidget)w)->chooseVar;
    int i, space, next_space;
    Widget prev = NULL, prevb;
    Widget form, title, bform, xit, rest, cancel;
    static XrmQuark SpaceQ = NULLQUARK;
    static XrmQuark ButtonQ = NULLQUARK;

    if ( !charQ ) {
	charQ = XrmStringToQuark("Char");
	ucharQ = XrmStringToQuark("UChar");
	shortQ = XrmStringToQuark("Short");
	ushortQ = XrmStringToQuark("UShort");
	intQ = XrmStringToQuark("Int");
	uintQ = XrmStringToQuark("UInt");
	floatQ = XrmStringToQuark("Float");
	doubleQ = XrmStringToQuark("Double");
    }

    if ( !SpaceQ  ) SpaceQ  = XrmStringToQuark("Space");
    if ( !ButtonQ ) ButtonQ = XrmStringToQuark("Button");

    form = XtNameToWidget(w, "form");
    if ( !form )
      form = XtVaCreateManagedWidget("form", formWidgetClass, w, NULL);

    title = XtNameToWidget(form, "title");
    if ( !title )
      title = XtVaCreateManagedWidget("title", labelWidgetClass, form, 
				      XtNlabel, cvp->title,
				      NULL);
    else
      XtVaSetValues(title, XtNlabel, cvp->title, NULL);

    prev = title;
    space = 0;
    for ( i = 0; i < cvp->choices.num; ++i ) {
	cvChoice *choice = &cvp->choices.ch[i];
	int n = findType(choice->type, choice->idx);

	next_space = 0;
	if ( n >= 0 ) {
	    choice->widget = (*cvTypeList[n].createf)(form, choice);
	    if ( cvTypeList[n].initf) (*cvTypeList[n].initf)(choice,True);
	} else if ( choice->type == SpaceQ ) {
	    next_space = choice->idx;
	} else if ( choice->type == ButtonQ ) {
	    if ( strcmp(choice->prompt, cvbEndButtons) ) {
		Widget bw;

		if ( strcmp(choice->prompt, cvbSameLine) ) {
		    Widget bprompt;

		    bform = XtNameToWidget(form, choice->prompt);
		    bform =
		      XtVaCreateManagedWidget(choice->prompt,
					      formWidgetClass, form, NULL);
		    bprompt = 
		      XtVaCreateManagedWidget("prompt",labelWidgetClass,bform, 
					      XtNlabel, choice->prompt, 
					      NULL);
		    choice->widget = bform;
		    prevb = bprompt;
		}

		bw = XtNameToWidget(bform, choice->name);
		if ( !bw )
		  bw = XtVaCreateManagedWidget(choice->name,commandWidgetClass,
					       bform, NULL);
		XtRemoveAllCallbacks(bw, XtNcallback);
		XtAddCallback(bw, XtNcallback, choice->changep, choice->var );
		XtVaSetValues(bw, XtNfromHoriz, prevb, NULL);
		prevb = bw;
	    }
	}

	if ( choice->widget ) {
	    XtVaSetValues(choice->widget, XtNfromVert, prev, NULL);
	    prev = choice->widget;
	    if ( space != 0 ) {
		int dist;
		XtVaGetValues(choice->widget, XtNvertDistance, &dist, NULL);
		XtVaSetValues(choice->widget, 
			      XtNvertDistance, dist+space, NULL);
	    }
	}
	space = next_space;
    }

    /* Create the Exit button. */
    xit = XtNameToWidget(form, "exit");
    if ( !xit )
      xit = XtVaCreateManagedWidget("exit", commandWidgetClass, form, NULL);
    XtVaSetValues(xit, XtNfromVert, prev, NULL );
    XtRemoveAllCallbacks(xit, XtNcallback);
    XtAddCallback(xit, XtNcallback, cvDone, w );
    prevb = xit;

    /* Now create any buttons specified in the choices list. */
    for ( i = 0; i < cvp->choices.num; ++i ) {
	cvChoice *choice = &cvp->choices.ch[i];
	
	if ( choice->type == ButtonQ && 
	    !strcmp(choice->prompt, cvbEndButtons) ) {
	    Widget bw = XtNameToWidget(form, choice->name);
	    if ( !bw )
	      bw = XtVaCreateManagedWidget(choice->name, commandWidgetClass, 
					   form, NULL);
	    XtVaSetValues(bw, 
			  XtNfromHoriz, prevb, XtNfromVert, prev, NULL );
	    XtRemoveAllCallbacks(bw, XtNcallback);
	    XtAddCallback(bw, XtNcallback, choice->changep, choice->var );
	    prevb = bw;
	}
    }

    /* Create the reset button. */
    rest = XtNameToWidget(form, "restore");
    if ( !rest )
      rest = XtVaCreateManagedWidget("restore",commandWidgetClass,form,NULL);
    XtVaSetValues(rest, XtNfromHoriz, prevb, XtNfromVert, prev, NULL );
    XtRemoveAllCallbacks(rest, XtNcallback);
    XtAddCallback(rest, XtNcallback, cvRestore, w );
    prevb = rest;

    /* Create the Cancel button. */
    cancel = XtNameToWidget(form, "cancel");
    if ( !cancel )
      cancel = XtVaCreateManagedWidget("cancel",commandWidgetClass,form,NULL);
    XtVaSetValues(cancel, XtNfromHoriz, prevb, XtNfromVert, prev, NULL );
    XtRemoveAllCallbacks(cancel, XtNcallback);
    XtAddCallback(cancel, XtNcallback, cvCancel, w );
    prevb = rest;
}

/* 
 * Find the "next" field which is an AsciiTextWidget.
 */

#include <X11/CompositeP.h>

Widget
  cvNextEditableField _IMPL((w, from), Widget w _AND Widget from )
{
    if ( XtIsSubclass(w, asciiTextWidgetClass) ) {
	return w;
    } else if ( XtIsComposite(w) ) {
	CompositeWidget cw = (CompositeWidget) w;
	int i;
	Boolean from_found = (from == 0);
	for ( i = 0; i < cw->composite.num_children; ++i ) {
	    Widget child = cw->composite.children[i];
	    Widget val;
	    if ( from_found ) {
		val = cvNextEditableField(child, 0);
		if ( val ) return val;
	    }
	    if ( child == from ) from_found = TRUE;
	}
    }

    return (!XtIsSubclass(XtParent(w), shellWidgetClass))? 
      cvNextEditableField(XtParent(w), w): (Widget)NULL;
}

void
  cvReinitValues _IMPL((cvw, new), ChooseVarWidget cvw _AND Boolean new )
{
    ChooseVarWidgetPart *cvp = &cvw->chooseVar;
    int i;

    for ( i = 0; i < cvp->choices.num; ++i ) {
	cvChoice *choice = &cvp->choices.ch[i];
	int n = findType(choice->type, choice->idx);
	
	if ( n >= 0 && cvTypeList[n].initf ) 
	  (*cvTypeList[n].initf)(choice,new);
    }
}

static void
  cvResetValues _IMPL((cvw), ChooseVarWidget cvw )
{
    ChooseVarWidgetPart *cvp = &cvw->chooseVar;
    int i;

    for ( i = 0; i < cvp->choices.num; ++i ) {
	cvChoice *choice = &cvp->choices.ch[i];
	int n = findType(choice->type, choice->idx);
	
	if ( n >= 0 && cvTypeList[n].resetf && 
	     (!cvTypeList[n].changedf || (*cvTypeList[n].changedf)(choice)) ) 
	  (*cvTypeList[n].resetf)(choice);
    }
    cvw->chooseVar.changed = False;
}

static void
  cvCheckChanged _IMPL((cvw), ChooseVarWidget cvw )
{
    ChooseVarWidgetPart *cvp = &cvw->chooseVar;
    int i;

    cvw->chooseVar.changed = False;
    for ( i = 0; i < cvp->choices.num; ++i ) {
	cvChoice *choice = &cvp->choices.ch[i];
	int n = findType(choice->type, choice->idx);
	
	if ( n >= 0 && cvTypeList[n].changedf && 
	     (*cvTypeList[n].changedf)(choice) ) {
	    cvw->chooseVar.changed = True;
	    break;
	}
    }
}

cvChoice *
getCVChoice _IMPL((w), Widget w )
{
    ChooseVarWidget cvw = findCVWidget(w);
    Widget sub;
    int i;

    for ( i = 0; i < cvw->chooseVar.choices.num; ++i ) {
	for ( sub = w; sub && sub != (Widget) cvw; sub = XtParent(sub) ) {
	    if ( sub == cvw->chooseVar.choices.ch[i].widget )
	      return &cvw->chooseVar.choices.ch[i];
	}
    }
    return NULL;
}

/*
 * This action takes two arguments, the widget name, and the callback 
 * name.  Without the second argument it defaults to "callback".
 */

static void
actCallback _IMPL((w, event, params, nparams),
		  Widget w _AND XEvent *event _AND 
		  String *params _AND Cardinal *nparams )
{
    Widget cvw = (Widget) findCVWidget(w);
    String wname = params[0];
    String cbname = XtNcallback;
    Widget cbw;

    if ( *nparams < 1 ) {
	fprintf(stderr, "Callback action requires widget name.\n");
	return;
    }

    if ( *nparams > 1 ) cbname = params[1];

    if ( cbw = XtNameToWidget(cvw, wname) )
      XtCallCallbacks( cbw, cbname, NULL );
}

void
  cvDone _IMPL((w, closure, call_data),
	       Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    Widget shell = (Widget) closure;
    
    cvCheckChanged(cvw);
    if ( ((ShellWidget) shell)->shell.popped_up ) XtPopdown(shell);
}

void
  cvRestore _IMPL((w, closure, call_data),
		  Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    Widget shell = (Widget) closure;

    cvResetValues(cvw);
}

void
  cvCancel _IMPL((w, closure, call_data),
		 Widget w _AND XtPointer closure _AND XtPointer call_data )
{
    ChooseVarWidget cvw = findCVWidget(w);
    Widget shell = (Widget) closure;

    if ( ((ShellWidget) shell)->shell.popped_up ) XtPopdown(shell);
    cvResetValues(cvw);
}

static void
actEditField _IMPL((w, ev, args, argc),
		   Widget w _AND XEvent *ev _AND String *args _AND int *argc )
{
    ChooseVarWidget cvw = findCVWidget(w);
    Widget source;

    if ( !cvw ) return;

    XtVaGetValues(w, XtNtextSource, &source, NULL);
    if ( !source ) return;

    XtOverrideTranslations(w, cvEditTrans);
    XtVaSetValues(source, XtNeditType, XawtextEdit, NULL);
    XtSetKeyboardFocus((Widget) cvw, w);
}

static void
actDoneField _IMPL((w, ev, args, argc),
		   Widget w _AND XEvent *ev _AND 
		   String *args _AND int *argc )
{
    ChooseVarWidget cvw = findCVWidget(w);
    cvChoice *choice = getCVChoice(w);
    Widget source;
    int n;
    
    if ( !cvw ) return;

    if ( choice ) {
	n = findType(choice->type, choice->idx);
	(*cvTypeList[n].changef)(w);
    }

    XtVaGetValues(w, XtNtextSource, &source, NULL);
    if ( !source ) return;

    XtOverrideTranslations(w, cvReadTrans);
    XtVaSetValues(source, XtNeditType, XawtextRead, NULL);
    XtSetKeyboardFocus((Widget) cvw, None);
}

static void
actEditNextField _IMPL((w, ev, args, argc),
		       Widget w _AND XEvent *ev _AND 
		       String *args _AND int *argc )
{
    ChooseVarWidget cvw = findCVWidget(w);
    Widget source;

    if ( !cvw ) return;

    XtVaGetValues(w, XtNtextSource, &source, NULL);
    if ( !source ) return;

    actDoneField(w, ev, args, argc);
    w = cvNextEditableField(XtParent(w), w);
    actEditField(w, ev, args, argc);
}

static XtActionsRec cvActions[] = { 
    {"cvEditField", actEditField},
    {"cvEditNextField", actEditNextField},
    {"cvDoneField", actDoneField},

    {"cvCallCallbacks", actCallback}
};
static Cardinal cvActionsCount = XtNumber(cvActions);

XtTranslations cvReadTrans = 0;
static char cvReadTransString[] = "#override \
     <Btn1Up>:     extend-end(PRIMARY, CUT_BUFFER0) cvEditField() \n\
Ctrl <Key> C:      cvCallCallbacks(form.cancel) \n\
Ctrl <Key> R:      cvCallCallbacks(form.restore) \n\
Ctrl <Key> S:      no-op(RingBell) \n\
     <Key> Return: cvCallCallbacks(form.exit) \n\
     <Key> Tab:    no-op(RingBell) \n\
     <FocusIn>:    display-caret(on) \n\
     <FocusOut>:   display-caret(off) \n\
";

XtTranslations cvEditTrans = 0;
static char cvEditTransString[] = "#override \
     <Btn1Up>:     extend-end(PRIMARY, CUT_BUFFER0) \n\
Ctrl <Key> C:      cvCallCallbacks(form.cancel) \n\
Ctrl <Key> R:      cvDoneField() cvCallCallbacks(form.restore) cvEditField()\n\
Ctrl <Key> S:      no-op(RingBell) \n\
     <Key> Return: cvDoneField() \n\
     <Key> Tab:    cvEditNextField() \n\
     <FocusIn>:    display-caret(on) \n\
     <FocusOut>:   display-caret(off) \n\
";

static void
ActionsInitialize _IMPL((w), Widget w)
{
    XtAppContext app_context = XtWidgetToApplicationContext(w);

    XtAppAddActions(app_context, cvActions, cvActionsCount);
    cvReadTrans = XtParseTranslationTable(cvReadTransString);
    cvEditTrans = XtParseTranslationTable(cvEditTransString);
}

Widget 
cvCreateAsciiTextWidget _IMPL((name, parent), String name _AND Widget parent )
{
    Widget entry;

    entry = XtVaCreateManagedWidget(name, asciiTextWidgetClass, parent, 
				    XtNuseStringInPlace, False,
				    XtNeditType, XawtextRead,
				    NULL);
    XtOverrideTranslations(entry, cvReadTrans);
    XtInstallAllAccelerators(entry, entry);

    return entry;
}


/* 
 * Boolean data entry.
 */

typedef struct {
    Boolean val;
    Widget yes, no;
} cvBooleanData;

static void
  ChangeBoolean _DCL((w, p), Widget w _AND cvChoice *p );
static void
  InitBoolean _DCL((p, new), cvChoice *p _AND Boolean new );
static void
  ResetBoolean _DCL((p), cvChoice *p );
static Boolean
  ChangedBoolean _DCL((p), cvChoice *p );

Widget
CreateBoolean _IMPL((w, p), Widget w _AND cvChoice *p )
{
    Widget form, label, yes, no;
    cvBooleanData *cv_data = p->cv_data;

    if ( !cv_data )
      cv_data = p->cv_data = XtNew(cvBooleanData);

    if ( !p->widget ) {
	form = p->widget =
	  XtVaCreateManagedWidget("boolean", formWidgetClass, w, NULL);
	label = XtVaCreateManagedWidget("prompt", labelWidgetClass, form,
					XtNlabel, p->prompt,
					NULL);
	yes = XtVaCreateManagedWidget("yes", commandWidgetClass, form, 
				      XtNfromHoriz, label,
				      NULL);

	no = XtVaCreateManagedWidget("no", commandWidgetClass, form,
				     XtNfromHoriz, yes,
				     NULL);
    } else {
	yes = XtNameToWidget(p->widget, "yes");
	no = XtNameToWidget(p->widget, "no");
    }

    XtRemoveAllCallbacks(yes, XtNcallback);
    XtAddCallback( yes, XtNcallback, ChangeBoolean, p );
    cv_data->yes = yes;

    XtRemoveAllCallbacks(no, XtNcallback);
    XtAddCallback( no, XtNcallback, ChangeBoolean, p );
    cv_data->no = no;

    return p->widget;
}

static void
ChangeBoolean _IMPL((w, p), Widget w _AND cvChoice *p )
{
    cvBooleanData *cv_data = p->cv_data;
    ChooseVarWidgetPart *cv = &((ChooseVarWidget) p->cv_widget)->chooseVar;
    Boolean *curr = (Boolean*)p->var;
    Boolean val = (cv_data->yes == w);

    *curr = val;
    XtVaSetValues(cv_data->yes,
		  XtNfont, val? cv->sel_font: cv->unsel_font, 
		  NULL);
    XtVaSetValues(cv_data->no, 
		  XtNfont, val? cv->unsel_font: cv->sel_font, 
		  NULL);
}

static void
InitBoolean _IMPL((p, new), cvChoice *p _AND Boolean new )
{
    cvBooleanData *cv_data = p->cv_data;
    Boolean *curr = (Boolean*)p->var;

    if ( new ) cv_data->val = *curr;
    ChangeBoolean(*curr? cv_data->yes: cv_data->no, p);
}

static void
ResetBoolean _IMPL((p), cvChoice *p )
{
    cvBooleanData *cv_data = p->cv_data;
    Boolean *curr = (Boolean*)p->var;
    
    *curr = cv_data->val;
    ChangeBoolean(cv_data->val? cv_data->yes: cv_data->no, p);
}

static Boolean
ChangedBoolean _IMPL((p), cvChoice *p )
{
    cvBooleanData *cv_data = p->cv_data;
    Boolean *curr = (Boolean*)p->var;

    return (*curr != cv_data->val);
}

/* 
 * Enum data entry.
 */

typedef struct {
    int val;
    int num;
    Widget *wids;
    XtPointer *vals;
} cvEnumData;

static void ChangeEnum _DCL((w, p), Widget w _AND cvChoice *p );
static void InitEnum _DCL((p, new), cvChoice *p _AND Boolean new );
static void ResetEnum _DCL((p), cvChoice *p );
static Boolean SetEnum _DCL((p), cvChoice *p );

Widget
CreateEnum _IMPL((w, p), Widget w _AND cvChoice *p )
{
    Widget form, prompt, prev, new;
    int i, j, num = 0;
    cvEnumData *cv_data = p->cv_data;

    if ( !cv_data ) {
	cv_data = p->cv_data = XtNew(cvEnumData);
	cv_data->wids = 0; 
	cv_data->vals = 0;
    }

    if ( !p->widget ) {
	p->widget = XtVaCreateManagedWidget("enum", formWidgetClass, w, NULL);
	prompt = XtVaCreateManagedWidget("prompt", labelWidgetClass, p->widget,
					 XtNlabel, p->prompt,
					 NULL);
    } else {
	prompt = XtNameToWidget(p->widget, "prompt");
    }
    form = p->widget;

    if ( p->args ) for ( num = 0; p->args[2*num]; ++num );
    cv_data->num = num;
    if ( num > 0 ) {
	cv_data->wids = 
	  (Widget *) XtRealloc(cv_data->wids, num * sizeof(Widget));
	cv_data->vals = 
	  (XtPointer) XtRealloc(cv_data->vals, num * sizeof(XtPointer));
    }

    prev = prompt;
    for ( i = j = 0; i < num; ++i ) {
	char name[10];
	sprintf(name, "enum%d", i);
	if ( !(new = XtNameToWidget(form, name)) )
	  new = XtVaCreateManagedWidget(name, commandWidgetClass, form, NULL);
	XtVaSetValues(new, XtNlabel, p->args[j++], XtNfromHoriz, prev, NULL);
	XtRemoveAllCallbacks(new, XtNcallback);
	XtAddCallback(new, XtNcallback, ChangeEnum, p );
	prev = new;

	cv_data->wids[i] = new;
	cv_data->vals[i] = p->args[j++];
    }

    return p->widget;
}

static void
ChangeEnum _IMPL((w, p), Widget w _AND cvChoice *p )
{
    cvEnumData *cv_data = p->cv_data;
    ChooseVarWidgetPart *cv = &((ChooseVarWidget) p->cv_widget)->chooseVar;
    XtPointer *curr = (XtPointer*)p->var;
    int i;

    for ( i = 0; i < cv_data->num; ++i ) {
	if ( w == cv_data->wids[i] ) {
	    *curr = cv_data->vals[i];
	    XtVaSetValues(w, XtNfont, cv->sel_font, NULL);
	} else {
	    XtVaSetValues(cv_data->wids[i], XtNfont, cv->unsel_font, NULL);
	}
    }
}

static void
InitEnum _IMPL((p, new), cvChoice *p _AND Boolean new )
{
    cvEnumData *cv_data = p->cv_data;
    XtPointer *curr = (XtPointer*)p->var;
    int i;

    for ( i = 0; i < cv_data->num; ++i ) {
	if ( *curr == cv_data->vals[i] ) {
	    if ( new ) cv_data->val = i;
	    ChangeEnum( cv_data->wids[i], p );
	}
    }
}

static void
ResetEnum _IMPL((p), cvChoice *p )
{
    cvEnumData *cv_data = p->cv_data;
    XtPointer *curr = (XtPointer*)p->var;
    int i;

    *(XtPointer *)p->var = cv_data->vals[cv_data->val];
    for ( i = 0; i < cv_data->num; ++i ) {
	if ( *curr == cv_data->vals[i] )
	  ChangeEnum( cv_data->wids[i], p );
    }
}

static Boolean
ChangedEnum _IMPL((p), cvChoice *p )
{
    cvEnumData *cv_data = p->cv_data;
    XtPointer *curr = (XtPointer*)p->var;

    return (*curr != cv_data->vals[cv_data->val]);
}

/* 
 * Integer data entry.
 */

typedef struct {
    union {
	int val;
	unsigned int uval;
    } u;
    int min, max;
} cvIntData;

static void ChangeInt _DCL((w), Widget w );
static void InitInt _DCL((p, new), cvChoice *p _AND Boolean new);
static void ResetInt  _DCL((p), cvChoice *p );
static Boolean ChangedInt _DCL((p), cvChoice *p );

Widget
CreateInt _IMPL((w, p), Widget w _AND cvChoice *p )
{
    Widget form, prompt, entry;
    cvIntData *cv_data = p->cv_data;

    if ( !cv_data )
      cv_data = p->cv_data = XtNew(cvIntData);
    cv_data->min = cv_data->max = 0;

    if ( !p->widget ) {
	form = p->widget = 
	  XtVaCreateManagedWidget("int", formWidgetClass, w, NULL);
	prompt = XtVaCreateManagedWidget("prompt", labelWidgetClass, form,
					 XtNlabel, p->prompt,
					 NULL);
	entry = cvCreateAsciiTextWidget("entry", form);
	XtVaSetValues(entry, XtNfromHoriz, prompt, NULL);
    }

    return p->widget;
}

static void
ChangeInt _IMPL((w), Widget w )
{
    cvChoice *p = getCVChoice(w);
    cvIntData *cv_data = p->cv_data;
    String str;
    char buf[15];

    XtVaGetValues(w, XtNstring, &str, NULL);
    if ( p->type == charQ ) {
	char *curr = (char *)p->var;
	*curr = atoi(str);
	sprintf(buf, "%d", *curr);
    } else if ( p->type == ucharQ ) {
	unsigned char *curr = (unsigned char *)p->var;
	*curr = atoi(str);
	sprintf(buf, "%u", *curr);
    } else if ( p->type == shortQ ) {
	short *curr = (short *)p->var;
	*curr = atoi(str);
	sprintf(buf, "%d", *curr);
    } else if ( p->type == ushortQ ) {
	unsigned short *curr = (unsigned short *)p->var;
	*curr = atoi(str);
	sprintf(buf, "%u", *curr);
    } else if ( p->type == intQ ) {
	int *curr = (int *)p->var;
	*curr = atoi(str);
	sprintf(buf, "%d", *curr);
    } else if ( p->type == uintQ ) {
	unsigned int *curr = (unsigned int *)p->var;
	*curr = atoi(str);
	sprintf(buf, "%u", *curr);
    }

    XtVaSetValues(w, XtNstring, buf, NULL);
}

static void
InitInt _IMPL((p, new), cvChoice *p _AND Boolean new )
{
    cvIntData *cv_data = p->cv_data;
    Widget entry = XtNameToWidget(p->widget, "entry");
    char buf[15];
    
    if ( p->type == charQ ) {
	char *curr = (char *)p->var;
	if ( new ) cv_data->u.val = *curr;
	sprintf(buf, "%d", *curr);
    } else if ( p->type == ucharQ ) {
	unsigned char *curr = (unsigned char *)p->var;
	if ( new ) cv_data->u.uval = *curr;
	sprintf(buf, "%u", *curr);
    } else if ( p->type == shortQ ) {
	short *curr = (short *)p->var;
	if ( new ) cv_data->u.val = *curr;
	sprintf(buf, "%d", *curr);
    } else if ( p->type == ushortQ ) {
	unsigned short *curr = (unsigned short *)p->var;
	if ( new ) cv_data->u.uval = *curr;
	sprintf(buf, "%u", *curr);
    } else if ( p->type == intQ ) {
	int *curr = (int *)p->var;
	if ( new ) cv_data->u.val = *curr;
	sprintf(buf, "%d", *curr);
    } else if ( p->type == uintQ ) {
	unsigned int *curr = (unsigned int *)p->var;
	if ( new ) cv_data->u.uval = *curr;
	sprintf(buf, "%u", *curr);
    } else {
	fprintf(stderr, "ResetInt(): Unrecognized integer type %s.\n",
		XrmQuarkToString(p->type));
	return;
    }

    if ( !entry ) {
	fprintf(stderr, "ResetInt(): Cannot find \"entry\" widget.\n");
	return;
    }
    XtVaSetValues(entry, XtNstring, buf, NULL);
}

static void
ResetInt _IMPL((p), cvChoice *p )
{
    cvIntData *cv_data = p->cv_data;
    Widget entry = XtNameToWidget(p->widget, "entry");
    char buf[15];

    if ( p->type == charQ ) {
	char *curr = (char *)p->var;
	*curr = cv_data->u.val;
	sprintf(buf, "%d", *curr);
    } else if ( p->type == ucharQ ) {
	unsigned char *curr = (unsigned char *)p->var;
	*curr = cv_data->u.uval;
	sprintf(buf, "%u", *curr);
    } else if ( p->type == shortQ ) {
	short *curr = (short *)p->var;
	*curr = cv_data->u.val;
	sprintf(buf, "%d", *curr);
    } else if ( p->type == ushortQ ) {
	unsigned short *curr = (unsigned short *)p->var;
	*curr = cv_data->u.uval;
	sprintf(buf, "%u", *curr);
    } else if ( p->type == intQ ) {
	int *curr = (int *)p->var;
	*curr = cv_data->u.val;
	sprintf(buf, "%d", *curr);
    } else if ( p->type == uintQ ) {
	unsigned int *curr = (unsigned int *)p->var;
	*curr = cv_data->u.uval;
	sprintf(buf, "%u", *curr);
    }

    if ( !entry ) return;
    XtVaSetValues(entry, XtNstring, buf, NULL);
}

static Boolean
ChangedInt _IMPL((p), cvChoice *p )
{
    cvIntData *cv_data = p->cv_data;

    if ( p->type == charQ ) {
	char *curr = (char *)p->var;
	return (*curr != cv_data->u.val);
    } else if ( p->type == ucharQ ) {
	unsigned char *curr = (unsigned char *)p->var;
	return (*curr != cv_data->u.uval);
    } else if ( p->type == shortQ ) {
	short *curr = (short *)p->var;
	return (*curr != cv_data->u.val);
    } else if ( p->type == ushortQ ) {
	unsigned short *curr = (unsigned short *)p->var;
	return (*curr != cv_data->u.uval);
    } else if ( p->type == intQ ) {
	int *curr = (int *)p->var;
	return (*curr != cv_data->u.val);
    } else if ( p->type == uintQ ) {
	unsigned int *curr = (unsigned int *)p->var;
	return (*curr != cv_data->u.uval);
    }
}

/* 
 * Float data entry.
 */

typedef struct {
    double val;
    double min, max;
} cvFloatData;

static void ChangeFloat _DCL((w), Widget w );
static void InitFloat _DCL((p, new), cvChoice *p _AND Boolean new );
static void ResetFloat _DCL((p), cvChoice *p );
static Boolean ChangedFloat _DCL((p), cvChoice *p );

Widget
CreateFloat _IMPL((w, p), Widget w _AND cvChoice *p )
{
    Widget form, prompt, entry;
    cvFloatData *cv_data = p->cv_data;

    if ( !cv_data )
      cv_data = p->cv_data = XtNew(cvFloatData);
    cv_data->min = cv_data->max = 0.0;

    if ( !p->widget ) {
	p->widget = form = 
	  XtVaCreateManagedWidget("float", formWidgetClass, w, NULL);
	prompt = XtVaCreateManagedWidget("prompt", labelWidgetClass, form,
					 XtNlabel, p->prompt,
					 NULL);
    
	entry = cvCreateAsciiTextWidget("entry", form);
	XtVaSetValues(entry, XtNfromHoriz, prompt, NULL);
    }

    return p->widget;
}

static void
ChangeFloat _IMPL((w), Widget w )
{
    cvChoice *p = getCVChoice(w);
    cvFloatData *cv_data = p->cv_data;
    float *curr = (float *)p->var;
    double atof();
    String str;
    char buf[15];

    XtVaGetValues(w, XtNstring, &str, NULL);
    if ( p->type == floatQ ) {
	float *curr = (float *)p->var;
	*curr = atof(str);
	sprintf(buf, "%g", *curr);
    } else {
	double *curr = (double *)p->var;
	*curr = atof(str);
	sprintf(buf, "%g", *curr);
    }
    XtVaSetValues(w, XtNstring, buf, NULL);
}

static void
InitFloat _IMPL((p, new), cvChoice *p _AND Boolean new )
{
    cvFloatData *cv_data = p->cv_data;
    Widget entry = XtNameToWidget(p->widget, "entry");
    char buf[15];

    if ( p->type == floatQ ) {
	float *curr = (float *)p->var;
	if ( new ) cv_data->val = *curr;
	sprintf(buf, "%g", *curr);
    } else {
	double *curr = (double *)p->var;
	if ( new ) cv_data->val = *curr;
	sprintf(buf, "%g", *curr);
    }

    if ( !entry ) {
	fprintf(stderr, "ResetFloat(): Cannot find \"entry\" widget.\n");
	return;
    }

    XtVaSetValues(entry, XtNstring, buf, NULL);
}

static void
ResetFloat _IMPL((p), cvChoice *p )
{
    cvFloatData *cv_data = p->cv_data;
    Widget entry = XtNameToWidget(p->widget, "entry");
    char buf[15];

    if ( p->type == floatQ ) {
	float *curr = (float *)p->var;
	*curr = cv_data->val;
	sprintf(buf, "%g", *curr);
    } else {
	double *curr = (double *)p->var;
	*curr = cv_data->val;
	sprintf(buf, "%g", *curr);
    }

    if ( !entry ) return;
    XtVaSetValues(entry, XtNstring, buf, NULL);
}

static Boolean
ChangedFloat _IMPL((p), cvChoice *p )
{
    cvFloatData *cv_data = p->cv_data;

    if ( p->type == floatQ ) {
	float *curr = (float *)p->var;
	return (*curr != cv_data->val);
    } else {
	double *curr = (double *)p->var;
	return (*curr != cv_data->val);
    }
}

/* 
 * String data entry.
 */

typedef struct {
    String val;
} cvStringData;

static void ChangeString _DCL((w), Widget w );
static void InitString _DCL((p, new), cvChoice *p _AND Boolean new);
static void ResetString _DCL((p), cvChoice *p );
static Boolean ChangedString _DCL((p), cvChoice *p );

Widget
CreateString _IMPL((w, p), Widget w _AND cvChoice *p )
{
    Widget form, prompt, entry;
    cvStringData *cv_data = p->cv_data;

    if ( !cv_data )
      cv_data = p->cv_data = XtNew(cvStringData);
    cv_data->val = 0;

    if ( !p->widget ) {
	form = p->widget =
	  XtVaCreateManagedWidget("string", formWidgetClass, w, NULL);
	prompt = XtVaCreateManagedWidget("prompt", labelWidgetClass, form,
					 XtNlabel, p->prompt,
					 NULL);
	entry = cvCreateAsciiTextWidget("entry", form);
	XtVaSetValues(entry, XtNfromHoriz, prompt, NULL);
    }

    return p->widget;
}

static void
ChangeString _IMPL((w), Widget w )
{
    cvChoice *p = getCVChoice(w);
    cvStringData *cv_data = p->cv_data;
    String *curr = (String *)p->var;
    String str;

    if ( *curr && *curr != cv_data->val ) free(*curr);

    XtVaGetValues(w, XtNstring, &str, NULL);
    *curr = strdup(str);
}

static void
InitString _IMPL((p, new), cvChoice *p _AND Boolean new )
{
    cvStringData *cv_data = p->cv_data;
    String *curr = (String *)p->var;
    Widget entry = XtNameToWidget(p->widget, "entry");
    String val;

    val = *curr;
    if ( new ) cv_data->val = val;
    
    if ( !entry ) {
	fprintf(stderr, "ResetString(): Cannot find \"entry\" widget.\n");
	return;
    }
    if ( val ) XtVaSetValues(entry, XtNstring, val, NULL);
}

static void
ResetString _IMPL((p), cvChoice *p )
{
    cvStringData *cv_data = p->cv_data;
    String *curr = (String *)p->var;
    Widget entry = XtNameToWidget(p->widget, "entry");

    if ( *curr && *curr != cv_data->val ) free(*curr);
    *curr = cv_data->val;

    if ( !entry ) return;
    XtVaSetValues(entry, XtNstring, *curr, NULL);
}

static Boolean
ChangedString _IMPL((p), cvChoice *p )
{
    cvStringData *cv_data = p->cv_data;
    String *curr = (String *)p->var;

    return (cv_data->val? 
	    strcmp(*curr, cv_data->val):
	    *curr == cv_data->val);
}

/* 
 * Register the standard set of type editors.
 */

static void
ClassInitialize _IMPL_VOID
{
    cvRegisterType("Boolean", 0, 
		   CreateBoolean, InitBoolean, ResetBoolean, 
		   NULL, ChangedBoolean );
    cvRegisterType("Enum", 0, 
		   CreateEnum, InitEnum, ResetEnum, 
		   NULL, ChangedEnum );
    cvRegisterType("Char", 0, 
		   CreateInt, InitInt, ResetInt, 
		   ChangeInt, ChangedInt );
    cvRegisterType("UChar", 0, 
		   CreateInt, InitInt, ResetInt, 
		   ChangeInt, ChangedInt );
    cvRegisterType("Short", 0, 
		   CreateInt, InitInt, ResetInt, 
		   ChangeInt, ChangedInt );
    cvRegisterType("UShort", 0, 
		   CreateInt, InitInt, ResetInt, 
		   ChangeInt, ChangedInt );
    cvRegisterType("Int", 0, 
		   CreateInt, InitInt, ResetInt, 
		   ChangeInt, ChangedInt );
    cvRegisterType("UInt", 0, 
		   CreateInt, InitInt, ResetInt, 
		   ChangeInt, ChangedInt );
    cvRegisterType("Float", 0, 
		   CreateFloat, InitFloat, ResetFloat, 
		   ChangeFloat, ChangedFloat );
    cvRegisterType("Double", 0, 
		   CreateFloat, InitFloat, ResetFloat, 
		   ChangeFloat, ChangedFloat );
    cvRegisterType("String", 0, 
		   CreateString, InitString, ResetString, 
		   ChangeString, ChangedString );
}


Boolean
ChooseVarModal _IMPL((shell, kind, placed),
		     ChooseVarWidget shell _AND 
		     XtGrabKind kind _AND Boolean placed )
{
    Widget popup = (Widget) shell;

    XtRealizeWidget(popup);
    XtInstallAllAccelerators(popup, popup);
    PlacePopup(popup, placed);
    XtPopup(popup, kind);

    while ( shell->shell.popped_up ) {
	XEvent event;
	XtAppNextEvent(XtWidgetToApplicationContext(popup),&event); 
        XtDispatchEvent(&event);
    }
    return shell->chooseVar.changed;
}
