/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: ImageActions.c,v 0.9.1.5 1991/12/16 20:57:27 leei Exp $";

#include <X11/copyright.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>

#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>

#include "EzMenu.h"
#include "ChooseVar.h"

#include "GenImageP.h"
#include "ImageP.h"
#include "ImViewportP.h"

#include "readImage.h"
#include "CMManage.h"

#include "proto.h"
#include "utils.h"
#include "debug.h"

#include <stdlib.h>
#include <regexp.h>
#include <math.h>

Widget
findApplicationShell _IMPL((w), Widget w )
{
    for ( ; w; w = w->core.parent )
      if ( XtIsSubclass(w, applicationShellWidgetClass) ) return w;

    return NULL;
}

Widget
findShell _IMPL((w), Widget w )
{
    for ( ; w; w = w->core.parent )
      if ( XtIsShell(w) ) return w;

    return NULL;
}

ImageViewportWidget
findImageViewport _IMPL((w, error), Widget w _AND Boolean error )
{
    for ( ; w; w = w->core.parent )
      if ( XtIsSubclass(w, imageViewportWidgetClass) )
	return (ImageViewportWidget) w;

    return NULL;
  }

ImageWidget
findImageWidget _IMPL((base, error), Widget base _AND Boolean error )
{
    Widget w;
    ImageViewportWidget vpw;

    for ( w = base; w; w = w->core.parent )
      if ( XtIsSubclass(w, imageWidgetClass) )
	return (ImageWidget) w;

    vpw = findImageViewport(base, False);
    if ( vpw ) return (ImageWidget) vpw->viewport.image;

    return NULL;
}

/*
 * These are the actions defined for an ImageViewport.
 */

static void 
actAbout _IMPL((ww, ev, str, card),
	       Widget ww _AND XEvent *ev _AND String *str _AND Cardinal *card )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    static Widget about = 0;
    Boolean placed = True;

    if ( !about ) {
	Widget form, dismiss;
	static XtPopdownIDRec popdown_ID;

	about = 
	  XtVaCreatePopupShell("about",transientShellWidgetClass,(Widget)w,
			       XtNtransientFor, (Widget)w,
			       NULL);
	form = XtVaCreateManagedWidget("form", formWidgetClass, about, NULL);
	XtVaCreateManagedWidget("text", asciiTextWidgetClass, form, 
				XtNtype, XawAsciiFile,
				XtNstring, ABOUT,
				XtNeditType, XawtextRead,
				NULL);
	dismiss =
	  XtVaCreateManagedWidget("dismiss", commandWidgetClass, form, NULL);
	popdown_ID.shell_widget = about;
	popdown_ID.enable_widget = (Widget) w;
	XtAddCallback(dismiss, XtNcallback, XtCallbackPopdown, &popdown_ID);
	XtRealizeWidget(about);
	placed = False;
    }
    
    PlacePopup(about, placed);
    XtPopup(about, XtGrabNone);
}

static void 
actCopyright _IMPL((ww, ev, str, card),
		   Widget ww _AND XEvent *ev _AND
		   String *str _AND Cardinal *card )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    static Widget cpyrght = 0;
    Boolean placed = True;

    if ( !cpyrght ) {
	Widget form, dismiss;
	extern char *imCopyright;
	static XtPopdownIDRec popdown_ID;

	cpyrght = 
	  XtVaCreatePopupShell("copyright",transientShellWidgetClass,(Widget)w,
			       XtNtransientFor, (Widget)w,
			       NULL);
	form = XtVaCreateManagedWidget("form", formWidgetClass, cpyrght, NULL);
	XtVaCreateManagedWidget("text", asciiTextWidgetClass, form, 
				XtNstring, imCopyright,
				XtNlength, strlen(imCopyright),
				XtNeditType, XawtextRead,
				XtNuseStringInPlace, True,
				NULL);
	dismiss =
	  XtVaCreateManagedWidget("dismiss", commandWidgetClass, form, NULL);
	popdown_ID.shell_widget = cpyrght;
	popdown_ID.enable_widget = (Widget) w;
	XtAddCallback(dismiss, XtNcallback, XtCallbackPopdown, &popdown_ID);
	XtRealizeWidget(cpyrght);
	placed = False;
    }
    
    PlacePopup(cpyrght, placed);
    XtPopup(cpyrght, XtGrabNone);
}

static void 
actShowMenuBar _IMPL((ww, ev, str, card),
		     Widget ww _AND XEvent *ev _AND 
		     String *str _AND Cardinal *card )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    Widget menu_bar;

    if ( !w ) return;
    
    menu_bar = w->viewport.menu_bar;
    if ( ev->xmotion.x < menu_bar->core.width &&
	 ev->xmotion.y < menu_bar->core.height ) {
	Mask mask = 0;
	XSetWindowAttributes attr;

	XtRealizeWidget(menu_bar);
	mask |= CWSaveUnder;
	attr.save_under = TRUE;
	XChangeWindowAttributes(XtDisplay(menu_bar),XtWindow(menu_bar), 
				mask, &attr);
	XMapRaised(XtDisplay(menu_bar), XtWindow(menu_bar));
    }
}

static void 
actHideMenuBar _IMPL((ww, ev, str, card),
		     Widget ww _AND XEvent *ev _AND 
		     String *str _AND Cardinal *card )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    Widget menu_bar;

    if ( !w ) return;
    
    menu_bar = w->viewport.menu_bar;
    XtUnmapWidget(menu_bar);
    XWithdrawWindow(XtDisplay(menu_bar), XtWindow(menu_bar),
		    XScreenNumberOfScreen(XtScreen(menu_bar)));
}

/* 
 * Setup one of the menuBar menus.  It modifies the menu named 
 * <menu_name> by inserting the <menu_descr> string *after* the 
 * <occurrence> occurrence of regexp <pat> in the pre-existing menu.
 * (-1 means the "last" occurrence). If <pat> is NULL, then the 
 * <menu_descr> string is placed at the beginning of the menu.  If 
 * <pat> is the empty string, then the end.  If <trunc> is True, then 
 * the menu is truncated beyond the inserted string.
 */

void
ImSetMenu _IMPL((ww, menu_name, menu_descr, pat, occurrence, trunc),
		Widget ww _AND
		char *menu_name _AND char *menu_descr _AND char *pat _AND 
		int occurrence _AND Boolean trunc )
{
    ImageViewportWidget w = findImageViewport(ww, True);
    Widget menu;
    String curr_menu, rest, new_menu;
    regexp *re = (pat? regcomp(pat): 0);

    /* Find the menu widget in the menu bar. */
    if ( !w ) return;
    menu = XtNameToWidget(w->viewport.menu_bar, menu_name);
    if ( !menu ) return;

    /* Get the current value of the menu_descr string. */
    XtVaGetValues(menu, XtNmenu, &curr_menu, NULL);

    /* If menu is currently empty, just replace with the given menu_descr. */
    if ( !curr_menu || !*curr_menu ) {
	XtVaSetValues(menu, XtNmenu, menu_descr, NULL);
	return;
    }

    /* Find the place to add the new menu. */
    if ( !pat || !occurrence ) 
      rest = curr_menu;
    else {
	rest = curr_menu;
	while ( occurrence-- != 0 ) {
	    if ( regexec(re, rest) )
	      rest = re->endp[0];
	    else
	      break;
	}
    }

    /* Create the new menu description: prefix + menu_descr + suffix. */
    new_menu = malloc(strlen(menu_descr) + (rest - curr_menu)
		      + (trunc? 0: strlen(rest)) + 2);

    /* Copy up to rest and NUL-terminate in case copied whole string. */
    strncpy(new_menu, curr_menu, rest-curr_menu);
    new_menu[rest-curr_menu] = '\0';
    strcat(new_menu, menu_descr);
    if ( !trunc ) strcat(new_menu, rest);

    /* Actually set the menu to its new incarnation. */
    XtVaSetValues(menu, XtNmenu, new_menu, NULL);
}

void actResizeToImage __((Widget ww, XEvent *ev, String *str, Cardinal *card));
void actPageLeft __((Widget ww, XEvent *ev, String *str, Cardinal *card));
void actPageRight __((Widget ww, XEvent *ev, String *str, Cardinal *card));
void actPageUp __((Widget ww, XEvent *ev, String *str, Cardinal *card));
void actPageDown __((Widget ww, XEvent *ev, String *str, Cardinal *card));

static void 
actRedraw _IMPL((ww, ev, str, card), 
		Widget ww _AND XEvent *ev _AND 
		String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);

    if ( !w ) return;
    
    ImRedraw(w, True);
}

static void 
actDelayRedraws _IMPL((ww, ev, str, card),
		      Widget ww _AND XEvent *ev _AND 
		      String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);

    w->image.redraws_delayed = True;
}

static void 
actGrid _IMPL((ww, ev, str, card),
	      Widget ww _AND XEvent *ev _AND String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);

    w->image.grid = !w->image.grid;
    ImRedraw(w, False);
}

static void
actResetZoom _IMPL((ww, ev, str, card),
		   Widget ww _AND XEvent *ev _AND String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);
    Arg argv[2]; int argc = 0;
    float one = 1.0;

    XtSetArg( argv[argc], XtNmagX, *(int*)&one ); ++argc;
    XtSetArg( argv[argc], XtNmagY, *(int*)&one ); ++argc;
    XtSetValues( (Widget)w, argv, argc );
}

static void
actZoomOut _IMPL((ww, ev, str, card),
		 Widget ww _AND XEvent *ev _AND 
		 String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);
    Arg argv[2]; int argc = 0;
    float new_mag_x = 0.5*w->image.mag_x;
    float new_mag_y = 0.5*w->image.mag_y;

    if ( MIN_SIZE/new_mag_x < w->core.width &&
	 MIN_SIZE/new_mag_y < w->core.height ) {
	XtSetArg( argv[argc], XtNmagX, *(int*)&new_mag_x ); ++argc;
	XtSetArg( argv[argc], XtNmagY, *(int*)&new_mag_y ); ++argc;
	XtSetValues( (Widget)w, argv, argc );
    }
}

static void
actZoomIn _IMPL((ww, ev, str, card), 
		Widget ww _AND XEvent *ev _AND
		String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);
    float new_mag_x = 2.0*w->image.mag_x;
    float new_mag_y = 2.0*w->image.mag_y;

    /* 
     * Only magnify if the parent of the window is large enough to 
     * make it significant.
     */
    if ( w->core.parent->core.width  >= MIN_PIXELS*w->image.mag_x &&
	  w->core.parent->core.height >= MIN_PIXELS*w->image.mag_y ) {
	XtVaSetValues((Widget) w, 
		      XtNmagX, *(int*)&new_mag_x,
		      XtNmagY, *(int*)&new_mag_y,
		      NULL );
    }
}

static void
actZoomBox _IMPL((ww, ev, str, card),
		 Widget ww _AND XEvent *ev _AND
		 String *str _AND Cardinal *card )
{
    int x1, y1, x2, y2;
    ImageWidget w = findImageWidget(ww, True);
    float mag_x = w->image.mag_x, mag_y = w->image.mag_y;
    float mag_ratio = mag_x/mag_y;
    Dimension p_width  = w->core.parent->core.width;
    Dimension p_height = w->core.parent->core.height;

    if ( ImSelectBox(w, &x1,&y1, &x2,&y2, ev) ) {
	mag_x = p_width /(float)(x2-x1);
	mag_y = p_height/(float)(y2-y1);

	if ( *card == 0 || strcmp(str[0],"nonconformal") ) {
	    if ( mag_x > mag_y*mag_ratio ) mag_x = mag_y*mag_ratio;
	    if ( mag_y > mag_x/mag_ratio ) mag_y = mag_x/mag_ratio;
	}

	x1 *= mag_x; /**/
	y1 *= mag_y; /**/

	XtVaSetValues( (Widget)w, 
		      XtNx, -x1, XtNy, -y1,
		      XtNmagX, *(int*)&mag_x, XtNmagY, *(int*)&mag_y,
		      NULL );
    }
}

static void
actFitToWindow _IMPL((ww, ev, str, card),
		     Widget ww _AND XEvent *ev _AND
		     String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);
    GenericDisplayParams disp = w->image.stack[0];
    float mag_x = disp->xparams.mag_x;
    float mag_y = disp->xparams.mag_y;
    float mag_ratio = mag_x/mag_y;
    Dimension width, height;
    Dimension p_width  = w->core.parent->core.width;
    Dimension p_height = w->core.parent->core.height;

    if ( w->image.stack_len ) {
	width = w->image.width;
	height = w->image.height;
	if ( !width || !height ) return;
	mag_x = p_width/(float)width;
	mag_y = p_height/(float)height;

	if ( *card == 0 || strcmp(str[0],"nonconformal") ) {
	    if ( mag_x > mag_y*mag_ratio ) mag_x = mag_y*mag_ratio;
	    if ( mag_y > mag_x/mag_ratio ) mag_y = mag_x/mag_ratio;
	}

	XtVaSetValues( (Widget)w, 
		      XtNx, 0, XtNy, 0,
		      XtNmagX, *(int*)&mag_x, XtNmagY, *(int*)&mag_y,
		      NULL );
    }
}

Widget image_fdlog = 0;

static void 
  actNewImage _IMPL((ww, ev, argv, argc),
		    Widget ww _AND XEvent *ev _AND
		    String *argv _AND Cardinal *argc )
{
    ImageWidget w = findImageWidget(ww, True);
    GenericImage image;
    char *fname;
    Boolean push = False;
    int i;

    if ( !image_fdlog ) {
	image_fdlog = makeFileDialog("file", findApplicationShell(ww));
	setFileDialogAcceptFunc(image_fdlog, acceptImageFile);
    }

    for ( i = 0; i < *argc; ++i, ++argv ) {
	if ( !strcmp(*argv,"push") ) push = True;
	else if ( !strcmp(*argv,"replace") ) push = False;
	else {
	    fprintf(stderr,"NewImage: Unrecognized argument \"%s\"\n",
		    *argv);
	}
    }

    fname = useFileDialog(image_fdlog, False);

    if ( fname ) {
	Display *dpy = XtDisplay((Widget)w);
	Window win = XtWindow((Widget)w);
	Cursor watch = XCreateFontCursor(dpy, XC_watch);

	XDefineCursor(dpy,win,watch);
	XSync(XtDisplay(w),False);
	if ( image = readImage(fname) ) {
	    /* 
	     * Only destroy if replacing and image is not same as one 
	     * already on top of the stack. 
	     */
	    if ( !push && w->image.stack_len ) 
	      ImPopImageStack(w, False, 
			      image != w->image.stack[0]->generic.image);
	    ImPushImageStack(w,image);
	}
	XUndefineCursor(dpy,win);
	XFreeCursor(dpy, watch);
    }
}

static void
  actPopImage _IMPL((ww, ev, argv, argc),
		    Widget ww _AND XEvent *ev _AND
		    String *argv _AND Cardinal *argc )
{
    ImageWidget w = findImageWidget(ww, True);
    Boolean destroy = True;
    Boolean all = False;

    if ( *argc != 0 ) {
	for ( ; *argc; --*argc, ++argv ) {
	    if ( !strcmp(*argv,"destroy") ) destroy = True;
	    else if ( !strcmp(*argv,"nodestroy") ) destroy = False;
	    else if ( !strcmp(*argv,"all") ) all = True;
	    else {
		fprintf(stderr,"PopImage: Unrecognized argument \"%s\"\n",
			*argv);
	    }
	}
    }
    
    if ( all ) {
	while ( w->image.stack_len ) 
	  ImPopImageStack(w,False,destroy);
	ImRedraw(w, False);
    } else {
	ImPopImageStack(w,True,destroy);
    }
}

static void
  actRotImage _IMPL((ww, ev, argv, argc),
		    Widget ww _AND XEvent *ev _AND
		    String *argv _AND Cardinal *argc )
{
    ImageWidget w = findImageWidget(ww, True);
    int n = 1;

    /* Takes one argument, num to rotate. */
    if ( *argc ) {
	n = atoi(*argv); ++argv; --argc;
    }
    
    ImRotImageStack(w,n,True);
}

static void
  actEditImParams _IMPL((ww, ev, argv, argc),
			Widget ww _AND XEvent *ev _AND
			String *argv _AND Cardinal *argc )
{
    ImageWidget w = findImageWidget(ww, True);
    GenericDisplayParams disp;

    if ( !w || w->image.stack_len <= 0 ) return;

    disp = w->image.stack[0];
    if ( w->image.stack_len > 0 ) ImEditParams(disp);
}

static void
  actEditStack _IMPL((ww, ev, argv, argc),
		     Widget ww _AND XEvent *ev _AND
		     String *argv _AND Cardinal *argc )
{
    ImageViewportWidget vpw = findImageViewport(ww, True);
    ImageWidget w = findImageWidget(ww, True);

    if ( !w->image.stack_editor ) 
      w->image.stack_editor = ImCreateStackEditor((Widget)w);

    ImUseStackEditor(w->image.stack_editor);
}

static void
  actEditParams _IMPL((ww, ev, argv, argc),
		      Widget ww _AND XEvent *ev _AND
		      String *argv _AND Cardinal *argc )
{
    ImageWidget w = findImageWidget(ww, True);
    float mag_x = w->image.mag_x, mag_y = w->image.mag_y;
    Boolean descr = w->image.descr;
    Boolean grid = w->image.grid;
    int displayN = w->image.displayN;
    Widget cv = w->image.cvWidget;
    Boolean placed = True;

    cvChoiceDescr Ch[6]; Cardinal Cn = 0;

    if ( !w ) return;

#if DEBUGGING
    cvInt(Ch[Cn], debug_mask, 0, "Debug:", NULL, NULL); ++Cn;
#endif
    cvFloat(Ch[Cn], mag_x,  0, "X Magnification:", NULL, NULL); ++Cn;
    cvFloat(Ch[Cn], mag_y,  0, "Y Magnification:", NULL, NULL); ++Cn;
    cvInt(Ch[Cn], displayN, 0, "Overlap:", NULL, NULL); ++Cn;
    cvBoolean(Ch[Cn], descr,   0, "Show Description?", NULL, NULL); ++Cn;
    cvBoolean(Ch[Cn], grid,    0, "Show Grid?", NULL, NULL); ++Cn;

    if ( !cv ) {
	placed = False;
	w->image.cvWidget = cv =
	  XtVaCreatePopupShell("imageChoices", chooseVarWidgetClass, ww, 
			       XtNcvTitle, "Edit Display Parameters",
			       NULL);
    }

    cvSetChoices(cv, Ch, Cn);

    if ( ChooseVarModal((ChooseVarWidget) cv, XtGrabExclusive, placed) ) {
	Arg argv[6];
	Cardinal argc = 0;

	if ( mag_x != w->image.mag_x ) 
	  XtSetArg( argv[argc], XtNmagX, *(int*)&mag_x ), ++argc;
	if ( mag_y != w->image.mag_y )
	  XtSetArg( argv[argc], XtNmagY, *(int*)&mag_y ), ++argc;
	if ( displayN != w->image.displayN )
	  XtSetArg( argv[argc], XtNdisplayN, displayN ), ++argc;
	if ( descr != w->image.descr )
	  XtSetArg( argv[argc], XtNdescription, descr ), ++argc;
	if ( grid != w->image.grid )
	  XtSetArg( argv[argc], XtNgrid, grid ), ++argc;
	if ( argc > 0 )
	  XtSetValues( ww, argv, argc );
    }
}

static void
  actSelectRegion _IMPL((ww, ev, str, card),
			Widget ww _AND XEvent *ev _AND
			String *str _AND Cardinal *card )
{
    int x1, y1, x2, y2;
    ImageWidget w = findImageWidget(ww, True);

    if ( ImSelectBox( w, &x1, &y1, &x2, &y2, ev ) ) {
	fprintf(stderr, "Selected region: %d,%d -> %d,%d\n", x1,y1, x2,y2);
    } else {
	fprintf(stderr, "Aborted select region.\n" );
    }
}

static void 
  actSelectLine _IMPL((ww, ev, str, card),
		      Widget ww _AND XEvent *ev _AND
		      String *str _AND Cardinal *card )
{
    int x1, y1, x2, y2;
    ImageWidget w = findImageWidget(ww, True);

    if ( ImSelectLine( w, &x1, &y1, &x2, &y2, ev ) ) {
	fprintf(stderr, "Selected Line: %d,%d -> %d,%d\n", x1,y1, x2,y2);
    } else {
	fprintf(stderr, "Aborted select line.\n" );
    }
}

static void 
  actSelectPoint _IMPL((ww, ev, str, card),
		       Widget ww _AND XEvent *ev _AND 
		       String *str _AND Cardinal *card )
{
    int x1, y1;
    ImageWidget w = findImageWidget(ww, True);

    if ( ImSelectPoint( w, &x1, &y1, ev ) ) {
	fprintf(stderr, "Selected Point: %d,%d\n", x1,y1);
    } else {
	fprintf(stderr, "Aborted select point.\n" );
    }
}

static void 
  actDescribePoint _IMPL((ww, ev, str, card),
			 Widget ww _AND XEvent *ev _AND
			 String *str _AND Cardinal *card )
{
    int i;
    int x, y;
    ImageWidget w = findImageWidget(ww, True);
    ImageWidgetPart *ipart = &w->image;
    char buf[512];

    if ( *card >= 2 ) {
	x = atoi(str[0]);
	y = atoi(str[1]);
    } else {
	ImSelectPoint( w, &x, &y, ev );
    }

    fprintf(stderr, "-- Point %d,%d\n", x, y);
    if ( ipart->stack_len ) {
	for ( i = MIN(ipart->displayN,ipart->stack_len) - 1; i >= 0; --i ) {
	    GenericDisplayParams params = ipart->stack[i];
	    buf[0] = '\0';
	    ImElementDescription(params->generic.image, x, y, buf,512);
	    fprintf(stderr, "%s", buf);
	}
    }
    fflush(stderr);
}

extern XrmQuark imStringQ;

static void 
  actSetValues _IMPL((ww, ev, str, card),
		     Widget ww _AND XEvent *ev _AND
		     String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);
    GenericDisplayParams disp = w->image.stack[0];

    if ( w->image.stack_len > 0 ) {
	GenericImage im = disp->generic.image;
	ImageClass cl = im->generic.class;
	XrmResourceList *xrmres = 
	  (XrmResourceList *) cl->generic.resources;
	Cardinal argc = 0;
	Arg argv[10];
	int i, j;

	for ( i = 0; i < *card/2; ++i ) {
	    XrmQuark nameQ = XrmStringToQuark(str[2*i]);
	    XrmQuark typeQ = -1;
	    XrmValue from, to;

	    /* Assumes a previously compiled resource list. */
	    for ( j = 0; j < cl->generic.num_resources; ++j ) {
		if ( nameQ == xrmres[j]->xrm_name ) {
		    typeQ = xrmres[j]->xrm_type;
		    break;
		}
	    }
	    if ( typeQ < 0 ) continue;

	    argv[argc].name = str[2*i];  
	    from.addr = str[2*i+1];
	    from.size = sizeof(String);
	    to.addr = 0; to.size = 0;
	    if ( ImConvert(imStringQ, &from, typeQ, &to) ) {
		if ( typeQ == imStringQ || to.size > sizeof(XtPointer) ) 
		  argv[argc].value = (int) to.addr;
		else switch (to.size) {
		  case 1:
		    argv[argc].value = *(char *)to.addr;
		    break;
		  case 2:
		    argv[argc].value = *(short *)to.addr;
		    break;
		  case 4:
		    argv[argc].value = *(int *)to.addr;
		    break;
		}
	    }
	    ++argc;
	}

	if ( argc ) ImSetValues(im, argv, argc);
    }
}

static void 
  actSetParams _IMPL((ww, ev, str, card),
		     Widget ww _AND XEvent *ev _AND 
		     String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);
    GenericDisplayParams disp = w->image.stack[0];
    if ( w->image.stack_len > 0 ) {
	GenericImage im = disp->generic.image;
	ImageClass cl = im->generic.class;
	XrmResourceList *xrmres = 
	  (XrmResourceList *) cl->generic.param_resources;
	Cardinal argc = 0;
	Arg argv[10];
	int i, j;

	for ( i = 0; i < *card/2; ++i ) {
	    XrmQuark nameQ = XrmStringToQuark(str[2*i]);
	    XrmQuark typeQ = -1;
	    XrmValue from, to;

	    /* Assumes a previously compiled resource list. */
	    for ( j = 0; j < cl->generic.num_param_resources; ++j ) {
		if ( nameQ == xrmres[j]->xrm_name ) {
		    typeQ = xrmres[j]->xrm_type;
		    break;
		}
	    }
	    if ( typeQ < 0 ) continue;

	    argv[argc].name = str[2*i]; 
	    from.addr = str[2*i+1];
	    from.size = sizeof(String);
	    to.addr = 0; to.size = 0;
	    if ( ImConvert(imStringQ, &from, typeQ, &to) ) {
		if ( typeQ == imStringQ || to.size > sizeof(XtPointer) ) 
		  argv[argc].value = (int) to.addr;
		else switch (to.size) {
		  case 1:
		    argv[argc].value = *(char *)to.addr;
		    break;
		  case 2:
		    argv[argc].value = *(short *)to.addr;
		    break;
		  case 4:
		    argv[argc].value = *(int *)to.addr;
		    break;
		}
	    }
	    ++argc;
	}

	if ( argc ) ImSetParams(disp, argv, argc);
    }
}

/* 
 * These routines implement a simple image action dispatching mechanism.
 * 
 * ImAddActions() is analogous to XtAppAddActions() and is used to 
 * register image action name/function pairs.
 * 
 * The defined action imAction(<actname>, <arg1>, <arg2>, ...) then 
 * dispatches to the image action named <actname> with the given 
 * arguments.  The imAction() action currently only dispatches against 
 * the top action on the stack.
 */

typedef struct {
    XrmQuark name;
    ImActionFunc func;
} imAction, *imActionList;

static imActionList im_actions;
static Cardinal num_im_actions, total_im_actions;

void
ImAddActions _IMPL((actions, num_actions),
		   ImActionList actions _AND Cardinal num_actions )
{
    imAction *curr;
    int i, j;

    /* Ensure that space is large enough. */
    if ( num_actions + num_im_actions > total_im_actions ) {
	total_im_actions = total_im_actions + num_actions + 10;
	im_actions = (imActionList) 
	  XtRealloc((char *)im_actions, total_im_actions * sizeof(imAction));
    }

    for ( i = 0; i < num_actions; ++i, ++actions ) {
	XrmQuark name = XrmStringToQuark(actions->name);
	
	/* Look for previous reference to name. */
	for ( j = 0, curr = im_actions; j < num_im_actions; ++j, ++curr ) {
	    if ( curr->name == name ) break;
	}
	if ( j == num_im_actions ) ++num_im_actions;
	
	/* Update/add reference to function. */
	curr->name = name;
	curr->func = actions->func;
    }
}

static void 
  actImAction _IMPL((ww, ev, str, card),
		    Widget ww _AND XEvent *ev _AND
		    String *str _AND Cardinal *card )
{
    ImageWidget w = findImageWidget(ww, True);
    GenericDisplayParams disp = w->image.stack[0];

    if ( w->image.stack_len > 0 ) {
	XrmQuark actname = XrmStringToQuark(str[0]);
	imAction *curr;
	int i;

	/* Search for the named action. */
	for ( i = 0, curr = im_actions; i < num_im_actions; ++i, ++curr )
	  if ( curr->name == actname ) break;

	if ( i >= num_im_actions ) {
	    char buf[BUFSIZ];
	    sprintf(buf, "imAction %s() not registered", str[0]);
	    XtAppWarning(XtWidgetToApplicationContext((Widget)w), buf);
	    return;
	}

	/* Now dispatch to registered action handler. */
	++str; --*card;
	(*curr->func)(disp, ww, ev, str, card);
    }
}

static XtActionsRec actions[] =
{
    /* {name, procedure}, */

    /* ImageViewport actions. */
    {"imAbout",			actAbout},
    {"imCopyright",		actCopyright},
    {"imShowMenuBar",		actShowMenuBar},
    {"imHideMenuBar",		actHideMenuBar},
    {"imPageLeft",		actPageLeft},
    {"imPageRight",		actPageRight},
    {"imPageUp",		actPageUp},
    {"imPageDown",		actPageDown},
    {"imResizeToImage", 	actResizeToImage},

    /* Image actions. */
    {"imRedraw",		actRedraw},
    {"imZoomIn",		actZoomIn},
    {"imZoomOut",		actZoomOut},
    {"imZoomBox",		actZoomBox},
    {"imFitToWindow",		actFitToWindow},
    {"imResetZoom",		actResetZoom},
    {"imToggleGrid",		actGrid},
    {"imDelayRedraws",		actDelayRedraws},
    {"imSelectLine",		actSelectLine},
    {"imSelectPoint",		actSelectPoint},
    {"imSelectRegion",		actSelectRegion},
    {"imPopImage",		actPopImage},
    {"imRotImage",		actRotImage},
    {"imNewImage",		actNewImage},
    {"imDescribePoint",		actDescribePoint},

    {"imEditStack",		actEditStack},
    {"imEditParams",		actEditParams},
    {"imSetValues",		actSetValues},
    {"imSetParams",		actSetParams},
    {"imAction",		actImAction},
    {"imEditImParams",		actEditImParams}
};

void
AddImageActions _IMPL((w), Widget w)
{
    XtAppAddActions( XtWidgetToApplicationContext(w), 
		     actions, XtNumber(actions) );
}
