/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

   $Id: gamma.h,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $

******************************************************************************/

#ifndef _gamma_h_
#define _gamma_h_

#include <X11/Intrinsic.h>

#include <vim/proto.h>

typedef enum { IHS=0, HexIHS, YIQ, RGB } ColorModel;

typedef enum { 
    gamDefault, gamFlat, gamEnhance, gamInverse, gamDull
  } WhichGamma;

typedef union {
    struct { float intens, hue, sat; } ihs;
    struct { float y, i, q; } yiq;
    struct { float r, g, b; } rgb;
} Color;

typedef struct {
    float in;
    float out;
} Anchor;

typedef struct {
    ColorModel model;
    Boolean smooth;
    Anchor anchors[5];
} GammaParams;

void defaultGamma __((GammaParams *params));
void chooseGamma __((GammaParams *params, WhichGamma which));
void chooseGammaModel __((GammaParams *params, ColorModel which));
void setGammaExp __((GammaParams *params, float expon));
void dimGamma __((GammaParams *params, float dimmer));
void brightenGamma __((GammaParams *params, float brighter));

float gammaCorrect __((GammaParams *params, XColor *in_out));

void SetAnchorGamma __((int idx, int in, int out));

void RGBtoFRGB __((XColor *in, Color *out));
void RGBtoIHS __((XColor *in, Color *out));
void RGBtoYIQ __((XColor *in, Color *out));
void RGBtoHexIHS __((XColor *in, Color *out));

void FRGBtoRGB __((Color *in, XColor *out));
void IHStoRGB __((Color *in, XColor *out));
void YIQtoRGB __((Color *in, XColor *out));
void HexIHStoRGB __((Color *in, XColor *out));

#endif /* _gamma_h_ */
