/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: gim_util.c,v 0.9.1.3 1991/12/16 21:04:30 leei Exp $";

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "gim_util.h"
#include "proto.h"
#include "strstr.h"

typedef int strip_arg;
#define strFRONT (1)
#define strEND   (2)

static char *
  strappend __(( char *a, char *b, Boolean dealloc ));
static char *
  strstrip __(( char *a, strip_arg which ));

char *
gim_type _IMPL((in), pFILE in )
{
    char line[82];
    char *p, *q;

    /* Peek at the first line. */
    pfgets(line, sizeof(line), in);
    pfungets(line, in);

    /* Now check type. */
    if ( (p=strchr(line,':')) == NULL ) return NULL;
    *p = '\0'; ++p;

    /* At this point, line is the property name and p is its value. */
    q = strstrip(line,strFRONT|strEND);
    p = strstrip(p,strFRONT|strEND);
    
    if ( strcmp("type",q) ) return NULL;
    return strdup(p);
}

ImPropList
gim_read_header _IMPL((in), pFILE in )
{
    int c;
    Cardinal len = 10;
    Cardinal propc = 0;
    ImPropList propv = (ImPropList) malloc( len * sizeof(ImProp) );
    char line[82];
    char *p, *q;
    Boolean dealloc;

    while ( (c = pfgetc(in)) != '\f' && c != EOF ) {
	pfungetc(c,in);
	pfgets(line,sizeof(line),in);
	if ( (p=strchr(line,':')) == NULL ) return 0;
	*p = '\0'; ++p;

	/* At this point, line is the property name and p is its value. */
	q = strstrip(line,strFRONT|strEND);
	p = strstrip(p,strFRONT);
	dealloc = False;

	/* Check for a continuation line. */
	while ( (c=pfgetc(in)), c != '\f' && isspace(c) ) {
	    char cont[82];
	    pfgets(cont,sizeof(cont),in);
	    p = strappend(p,cont,dealloc);
	    dealloc = True;
	}
	strstrip(p,strEND);
	pfungetc(c,in);

	/* Expand if necessary. */
	if ( propc == len-1 ) {
	    len += 10;
	    propv = (ImPropList) realloc( propv, len * sizeof(ImProp) );
	}

	ImSetProp( propv[propc], strdup(q), XtRString, strdup(p) ); ++propc;
    }

    ImSetProp( propv[propc], NULL, NULL, NULL );
    return propv;
}

caddr_t
gim_get_prop _IMPL((propv, name), ImPropList propv _AND String name )
{
    for ( ; propv->name; ++propv )
      if ( !strcmp(name,propv->name) )
	return propv->value.addr;
    return 0;
}

Cardinal
gim_numprops _IMPL((propv), ImPropList propv )
{
    Cardinal num = 0;
    for ( ; propv->name; ++propv, ++num );
    return num;
}

static char *
strappend _IMPL((a, b, dealloc), char *a _AND char *b _AND Boolean dealloc )
{
    if ( !a ) {
	return strdup(b);
    } else {
	int len = strlen(a) + strlen(b) + 1;
	char *new = (char *) malloc(len);
	strcpy(new,a);
	strcat(new,b);
	if ( dealloc ) free(a);
	return new;
    }
}

static char *
strstrip _IMPL((p, which), char *p _AND strip_arg which )
{
    char *q;

    if ( which & strFRONT ) {
	while ( *p && isspace(*p) ) ++p;
    }
    
    if ( which & strEND ) {
	q = p + strlen(p) - 1;
	while ( q > p && isspace(*q) ) *q-- = '\0';
    }

    return p;
}
