/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: pgm_read.c,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $";

/* From pbm.c:
 *
 * portable bit map (pbm) format images
 *
 * jim frost 09.27.89
 *
 * patched by W. David Higgins (wdh@mkt.csd.harris.com) to support
 * raw-format PBM files.
 *
 * patched by Ian MacPhedran (macphed@dvinci.usask.ca) to support
 * PGM and PPM files (03-July-1990)
 */

#include <vim/proto.h>
#include <vim/pFile.h>

#include <vim/ByteImageP.h>

#define BADREAD    0 /* read error */
#define NOTPBM     1 /* not a pbm file */
#define PBMNORMAL  2 /* pbm normal type file */
#define PBMCOMPACT 3 /* pbm compacty type file */
#define PBMRAWBITS 4 /* pbm raw bits type file */
#define PGMNORMAL  5 /* pgm normal type file */
#define PGMRAWBITS 6 /* pgm raw bytes type file */
#define PPMNORMAL  7 /* ppm normal type file */
#define PPMRAWBITS 8 /* ppm raw bytes type file */

GenericImage
pgm_read _IMPL((in, ident), pFILE in _AND char *ident )
{
    ByteImage im;
    char buf[3];
    int pgm_type;
    byte         *destptr;
    short curmax, maxval;
    int numbytes;
    int y;
    Arg argv[2]; int argc;
    declare_ByteImage(pgm);

    /* Determine the type of the PGM image. */
    buf[0] = pfgetc(in);
    buf[1] = pfgetc(in);
    buf[2] = '\0';
    pgm_type = ((!strncmp("P2",buf,2))? PGMNORMAL:
		(!strncmp("P5",buf,2))? PGMRAWBITS: 
		NOTPBM);

    pfgetuhint(&pgm_width,in);
    pfgetuhint(&pgm_height,in);
    pfgethint(&maxval,in); if ( !maxval ) maxval = 255;
    pfgets(NULL,-1,in);

    for (pgm_depth=0,curmax=1;maxval>curmax;pgm_depth++) curmax=curmax*2;

    argc = 0;
    XtSetArg( argv[argc], ImNdepth, pgm_depth ); ++argc;
    im = (ByteImage) 
      ImCreateImage( ident, byteImageClass, pgm_width, pgm_height, argv,argc );
    access_ByteImage(im,pgm);

    /* Initialize color map to a ramp. */
    for (y = 0; y <= maxval; y++) {
	pgm_cmap->red  [y] = (y * 0xffff)/maxval;
	pgm_cmap->green[y] = (y * 0xffff)/maxval;
	pgm_cmap->blue [y] = (y * 0xffff)/maxval;
    }
    pgm_cmap->used = maxval+1;

    switch ( pgm_type ) {
      case PGMRAWBITS:
	numbytes= pgm_width * pgm_height;
	for (y = 0, destptr = pgm_data; y < numbytes; y++, destptr++) {
	    if ( (*destptr = pfgetc(in)) < 0 ) {
		fprintf(stderr,"%s: Short image (%d of %d)\n", 
			ident, y, numbytes);
		goto done;
	    }
	    ++pgm_cmap->count[*destptr];
	}
	break;
      case PGMNORMAL:
	numbytes = pgm_width * pgm_height;
	for (y = 0, destptr = pgm_data; y < numbytes; y++, destptr++) {
	    if ( !pfgetucint(destptr,in) ) {
		fprintf(stderr,"%s: Short image (%d of %d)\n", 
			ident, y, numbytes);
		goto done;
	    }
	    ++pgm_cmap->count[*destptr];
	}
	break;
      default:
	fprintf(stderr,"%s: Unrecognized PGM format\n",ident);
	ImDestroyImage((GenericImage)im);
	return 0;
    }

  done:
    end_ByteImage(im,pgm);
    return (GenericImage)im;
}
