/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: ppm_read.c,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $";

/* From pbm.c:
 *
 * portable bit map (pbm) format images
 *
 * jim frost 09.27.89
 *
 * patched by W. David Higgins (wdh@mkt.csd.harris.com) to support
 * raw-format PBM files.
 *
 * patched by Ian MacPhedran (macphed@dvinci.usask.ca) to support
 * PGM and PPM files (03-July-1990)
 */

#include <stdlib.h>

#include <vim/pFile.h>
#include <vim/hash.h>
#include <vim/proto.h>

#include <RGBImageP.h>
#include <ByteImageP.h>

#define BADREAD    0 /* read error */
#define NOTPBM     1 /* not a pbm file */
#define PBMNORMAL  2 /* pbm normal type file */
#define PBMCOMPACT 3 /* pbm compacty type file */
#define PBMRAWBITS 4 /* pbm raw bits type file */
#define PGMNORMAL  5 /* pgm normal type file */
#define PGMRAWBITS 6 /* pgm raw bytes type file */
#define PPMNORMAL  7 /* ppm normal type file */
#define PPMRAWBITS 8 /* ppm raw bytes type file */

GenericImage
ppm_read _IMPL((in, ident), pFILE in _AND char *ident )
{
    GenericImage im;
    char buf[3];
    int ppm_type;
    short src;
    short curmax, maxval;
    int size;
    int x, y;
    Dimension width, height;
    int depth;
    byte *data;
    Arg argv[2]; int argc;
    hash_list chl;

    /* Determine the type of the PPM image. */
    buf[0] = pfgetc(in);
    buf[1] = pfgetc(in);
    buf[2] = '\0';
    ppm_type = ((!strncmp("P3",buf,2))? PPMNORMAL:
		(!strncmp("P6",buf,2))? PPMRAWBITS: 
		NOTPBM);

    pfgetuhint(&width,in);
    pfgetuhint(&height,in);
    pfgethint(&maxval,in); if ( !maxval ) maxval = 255;
    pfgets(NULL,-1,in); /**/

    size = 3*height*width;
    chl = init_hash();

    for ( depth=0, curmax=1; maxval>curmax; depth++) curmax *= 2;
    data = (byte *) malloc( size );

    switch ( ppm_type ) {

      case PPMRAWBITS:
	for (y=0; y<size; ) {
	    if ((src = pfgetc(in)) == EOF) {
		fprintf(stderr,"%s: Short image (%d of %d)\n", ident, y, size);
		break;
	    } else
	      data[y++] = src;
	    if ((src = pfgetc(in)) == EOF) {
		fprintf(stderr,"%s: Short image (%d of %d)\n", ident, y, size);
		break;
	    } else
	      data[y++] = src;
	    if ((src = pfgetc(in)) == EOF) {
		fprintf(stderr,"%s: Short image (%d of %d)\n", ident, y, size);
		break;
	    } else
	      data[y++] = src;
	    add_hash(chl, data[y-3], data[y-2], data[y-1]);
	}
	break;

      case PPMNORMAL:
	for (y=0; y<size; ) {
	    if ( !pfgethint(&src,in) ) {
		fprintf(stderr,"%s: Short image (%d of %d)\n", ident, y, size);
		break;
	    }
	    else
	      data[y++] = src;
	    if ( !pfgethint(&src,in) ) {
		fprintf(stderr,"%s: Short image (%d of %d)\n", ident, y, size);
		break;
	    }
	    else
	      data[y++] = src;
	    if ( !pfgethint(&src,in) ) {
		fprintf(stderr,"%s: Short image (%d of %d)\n", ident, y, size);
		break;
	    }
	    else
	      data[y++] = src;
	    add_hash(chl, data[y-3], data[y-2], data[y-1]);
	}
	break;

      default:
	fprintf(stderr,"%s: Unrecognized PPM format\n",ident);
	ImDestroyImage((GenericImage)im);
	return 0;
    }

    y = count_hash(chl); /* Find number of colours in table */

    /* Choose whether or not this is a ByteImage or an RGBImage. */
    if ( y <= 256 ) {
	ByteImage bim;
	byte *destline;
	declare_ByteImage(ppm);

	for (depth=0, curmax=1; y>curmax; depth++) curmax *= 2;
	argc = 0;
	XtSetArg( argv[argc], ImNdepth, depth ); ++argc;
	im = ImCreateImage(ident, byteImageClass, width, height, argv,argc );
	bim = (ByteImage) im;
	access_ByteImage(bim,ppm);

	ppm_cmap->used = y;
	fill_map(chl,ppm_cmap,maxval);
	destline = ppm_data;
	for (x=0; x<size; ) {
	    byte red = data[x++];
	    byte green = data[x++];
	    byte blue = data[x++];
	    *destline = y = find_hash( chl, red, green, blue);
	    ++ppm_cmap->count[y];
	    destline += ppm_pixlen;
	}
	free(data);
    } else {
	argc = 0;
	XtSetArg( argv[argc], ImNdepth, depth ); ++argc;
	XtSetArg( argv[argc], ImNdata,  data  ); ++argc;
	im = ImCreateImage(ident, rgbImageClass, width,height, argv,argc );
    }
	
    free_hash( chl );

    return (GenericImage)im;
}
