#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>

#include "gamma.h"

typedef enum { IHS, HexconeIHS, YIQ } ColorModel;

static void tgamQuit( Widget, XEvent*, String*, Cardinal*);
static void tgamDraw( Widget, XEvent*, String*, Cardinal*);
static void tgamSetRange( Widget, XEvent*, String*, Cardinal*);
static void tgamRemodel( Widget, XEvent*, String*, Cardinal*);

static XtActionsRec actions[] =
{
    /* {name, procedure}, */
    {"tgamDraw", tgamDraw},
    {"tgamSetRange", tgamSetRange},
    {"tgamRemodel", tgamRemodel},
    {"tgamQuit", tgamQuit}
};

static char *coreTrans = "#augment \n\
   c<Key>c:		tgamQuit()\n\
    <Key>q:		tgamQuit()\n\
    <Btn1Down>:		tgamSetRange()\n\
    <Btn2Down>:		tgamRemodel()\n\
    <Expose>:		tgamDraw()\n\
";

XtAppContext app_context;
Widget toplevel;

#define STR_W (8)
#define STR_H (32)
#define STR_N (32)

main(argc, argv)
Cardinal argc;
String *argv;
{
    toplevel = XtAppInitialize(&app_context, "TGamma",
			       NULL, 0, &argc, argv, NULL, 
			       NULL, 0);
    XtAppAddActions( app_context, actions, XtNumber(actions) );

    XtVaCreateManagedWidget( "core", coreWidgetClass, toplevel,
			     XtNwidth, 32 + STR_W*STR_N,
			     XtNheight, 32 + STR_H*3,
			     XtNtranslations, 
			       XtParseTranslationTable(coreTrans),
			     NULL );

    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);
}

static void 
tgamQuit( Widget w, XEvent *ev, String *str, Cardinal *card )
{
    exit(0);
}

static ColorModel which_model = YIQ;
static float range1 = 0.5;
static float range2 = 0.5;
static float range3 = 0.5;

static void 
tgamDraw( Widget w, XEvent *ev, String *str, Cardinal *card )
{
    switch ( which_model ) {
      case YIQ:
	DrawYIQStrips(XtDisplay(w), XtWindow(w),
		      DefaultColormapOfScreen(XtScreen(w)) );
	break;
      case IHS:
	DrawIHSStrips(XtDisplay(w), XtWindow(w),
		      DefaultColormapOfScreen(XtScreen(w)) );
	break;
      case HexconeIHS:
	DrawHexconeIHSStrips(XtDisplay(w), XtWindow(w),
			     DefaultColormapOfScreen(XtScreen(w)) );
	break;
    }
}

static void 
tgamSetRange( Widget w, XEvent *ev, String *str, Cardinal *card )
{
    int which = (ev->xbutton.y - 16)/STR_H;
    float nrange = (ev->xbutton.x - 16)/((float) STR_W*STR_N);

    nrange = nrange<0? 0: (nrange>1? 1: nrange);
    switch ( which ) {
      case 0: range1 = nrange; break;
      case 1: range2 = nrange; break;
      case 2: range3 = nrange; break;
      default: return;
    }
    XClearArea( XtDisplay(w), XtWindow(w), 0, 0, 0, 0, True );
}

static void 
tgamRemodel( Widget w, XEvent *ev, String *str, Cardinal *card )
{
    switch ( which_model ) {
      case YIQ: which_model = IHS; break;
      case IHS: which_model = HexconeIHS; break;
      case HexconeIHS: which_model = YIQ; break;
    }
    range1 = range2 = range3 = 0.5;
    XClearArea( XtDisplay(w), XtWindow(w), 0, 0, 0, 0, True );
}

DrawStrip( Display *disp, Drawable win, int x, int y, Pixel *pix )
{
    static GC gc;
    XGCValues gcvals;
    int i;

    if ( gc == 0 ) gc = XCreateGC(disp,win,0,0);
    for ( i = x; i < x+STR_W*STR_N; i+=STR_W, ++pix ) {
	gcvals.foreground = *pix;
	XChangeGC( disp, gc, GCForeground, &gcvals );
	XFillRectangle( disp, win, gc, i, y, STR_W, STR_H );
    }
}

const float y_min =  0.00, y_max =  1.00;
const float i_min = -0.60, i_max =  0.60;
const float q_min = -0.52, q_max =  0.52;

DrawYIQStrips( Display *disp, Drawable win, Colormap cmap )
{
    Color yiq; XColor xcolor;
    Pixel pixels[STR_N*3];
    int i;

    xcolor.flags = DoRed | DoGreen | DoBlue;
    
    yiq.yiq.y = y_min + range1*(y_max-y_min);
    yiq.yiq.i = i_min + range2*(i_max-i_min);
    yiq.yiq.q = q_min + range3*(q_max-q_min);
    for ( i = 0, yiq.yiq.y = y_min; 
	  i < STR_N; 
	  ++i, yiq.yiq.y += (y_max-y_min)/STR_N ) {
	YIQtoRGB(&yiq,&xcolor);
	XAllocColor(disp,cmap,&xcolor);
	pixels[i] = xcolor.pixel;
    }
    DrawStrip( disp, win, 16, 16, pixels );

    yiq.yiq.y = y_min + range1*(y_max-y_min);
    yiq.yiq.i = i_min + range2*(i_max-i_min);
    yiq.yiq.q = q_min + range3*(q_max-q_min);
    for ( i = 0, yiq.yiq.i = i_min; 
	  i < STR_N; 
	  ++i, yiq.yiq.i += (i_max-i_min)/STR_N ) {
	YIQtoRGB(&yiq,&xcolor);
	XAllocColor(disp,cmap,&xcolor);
	pixels[i+STR_N] = xcolor.pixel;
    }
    DrawStrip( disp, win, 16, 16+STR_H, pixels+STR_N );

    yiq.yiq.y = y_min + range1*(y_max-y_min);
    yiq.yiq.i = i_min + range2*(i_max-i_min);
    yiq.yiq.q = q_min + range3*(q_max-q_min);
    for ( i = 0, yiq.yiq.q = q_min; 
	  i < STR_N; 
	  ++i, yiq.yiq.q += (q_max-q_min)/STR_N ) {
	YIQtoRGB(&yiq,&xcolor);
	XAllocColor(disp,cmap,&xcolor);
	pixels[i+2*STR_N] = xcolor.pixel;
    }
    DrawStrip( disp, win, 16, 16+2*STR_H, pixels+2*STR_N );

    XFreeColors( disp, cmap, pixels, 3*STR_N, 0 );
}

const float ii_min =  0.00, ii_max =  1.72;
const float ih_min =  0.00, ih_max =  1.00;
const float is_min =  0.00, is_max =  1.00;

DrawIHSStrips( Display *disp, Drawable win, Colormap cmap )
{
    Color ihs; XColor xcolor;
    Pixel pixels[STR_N*3];
    int i;

    xcolor.flags = DoRed | DoGreen | DoBlue;

    ihs.ihs.intens = ii_min + range1*(ii_max-ii_min);
    ihs.ihs.hue = ih_min + range2*(ih_max-ih_min);
    ihs.ihs.sat = is_min + range3*(is_max-is_min);
    for ( i = 0, ihs.ihs.intens = ii_min;
	  i < STR_N; 
	  ++i, ihs.ihs.intens += (ii_max-ii_min)/STR_N ) {
	IHStoRGB(&ihs,&xcolor);
	XAllocColor(disp,cmap,&xcolor);
	pixels[i] = xcolor.pixel;
    }
    DrawStrip( disp, win, 16, 16, pixels );

    ihs.ihs.intens = ii_min + range1*(ii_max-ii_min);
    ihs.ihs.hue = ih_min + range2*(ih_max-ih_min);
    ihs.ihs.sat = is_min + range3*(is_max-is_min);
    for ( i = 0, ihs.ihs.hue = ih_min;
	  i < STR_N; 
	  ++i, ihs.ihs.hue += (ih_max-ih_min)/STR_N ) {
	IHStoRGB(&ihs,&xcolor);
	XAllocColor(disp,cmap,&xcolor);
	pixels[i+STR_N] = xcolor.pixel;
    }
    DrawStrip( disp, win, 16, 16+STR_H, pixels+STR_N );

    ihs.ihs.intens = ii_min + range1*(ii_max-ii_min);
    ihs.ihs.hue = ih_min + range2*(ih_max-ih_min);
    ihs.ihs.sat = is_min + range3*(is_max-is_min);
    for ( i = 0, ihs.ihs.sat = is_min;
	  i < STR_N; 
	  ++i, ihs.ihs.sat += (is_max-is_min)/STR_N ) {
	IHStoRGB(&ihs,&xcolor);
	XAllocColor(disp,cmap,&xcolor);
	pixels[i+2*STR_N] = xcolor.pixel;
    }
    DrawStrip( disp, win, 16, 16+2*STR_H, pixels+2*STR_N );

    XFreeColors( disp, cmap, pixels, 3*STR_N, 0 );
}

const float hi_min =  0.00, hi_max =  1.00;
const float hh_min =  0.00, hh_max =  1.00;
const float hs_min =  0.00, hs_max =  1.00;

DrawHexconeIHSStrips( Display *disp, Drawable win, Colormap cmap )
{
    Color ihs; XColor xcolor;
    Pixel pixels[STR_N*3];
    int i;

    xcolor.flags = DoRed | DoGreen | DoBlue;

    ihs.ihs.intens = hi_min + range1*(hi_max-hi_min);
    ihs.ihs.hue = hh_min + range2*(hh_max-hh_min);
    ihs.ihs.sat = hs_min + range3*(hs_max-hs_min);
    for ( i = 0, ihs.ihs.intens = hi_min; 
	  i < STR_N; 
	  ++i, ihs.ihs.intens += (hi_max-hi_min)/STR_N ) {
	HexIHStoRGB(&ihs,&xcolor);
	XAllocColor(disp,cmap,&xcolor);
	pixels[i] = xcolor.pixel;
    }
    DrawStrip( disp, win, 16, 16, pixels );

    ihs.ihs.intens = hi_min + range1*(hi_max-hi_min);
    ihs.ihs.hue = hh_min + range2*(hh_max-hh_min);
    ihs.ihs.sat = hs_min + range3*(hs_max-hs_min);
    for ( i = 0, ihs.ihs.hue = hh_min; 
	  i < STR_N; 
	  ++i, ihs.ihs.hue += (hh_max-hh_min)/STR_N ) {
	HexIHStoRGB(&ihs,&xcolor);
	XAllocColor(disp,cmap,&xcolor);
	pixels[i+STR_N] = xcolor.pixel;
    }
    DrawStrip( disp, win, 16, 16+STR_H, pixels+STR_N );

    ihs.ihs.intens = hi_min + range1*(hi_max-hi_min);
    ihs.ihs.hue = hh_min + range2*(hh_max-hh_min);
    ihs.ihs.sat = hs_min + range3*(hs_max-hs_min);
    for ( i = 0, ihs.ihs.sat = hs_min; 
	  i < STR_N; 
	  ++i, ihs.ihs.sat += (hs_max-hs_min)/STR_N ) {
	HexIHStoRGB(&ihs,&xcolor);
	XAllocColor(disp,cmap,&xcolor);
	pixels[i+2*STR_N] = xcolor.pixel;
    }
    DrawStrip( disp, win, 16, 16+2*STR_H, pixels+2*STR_N );

    XFreeColors( disp, cmap, pixels, 3*STR_N, 0 );
}
